/******************************************************************************
 * $Id: main.c,v 1.67 2000/07/22 06:16:24 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: "hypersrc" is a program for browsing source code using tags.
 ******************************************************************************/

#define HYPERSRC_DEFINE_GLOBALS
#include "header.h"

/******************************************************************************
 * Main.
 ******************************************************************************/
int main( int argc, char* argv[] )
{
  /*
   * Initialize globals.
   */
   InitGlobals();

  /*
   * Parse command line (after globals are initialized).
   */
   ParseCommandLine( argc, argv );

  /*
   * Initialize GTK (necessary before threading can occur).
   */
   gtk_init( &argc, &argv );

  /*
   * Even though GLIB thread functions are not used,
   * calling this seems necessary (otherwise, the tags processing
   * has been known to erratically hang, evidently, at a GLIB call).
   */
#ifdef HYPERSRC_THREADS
   if ( ! g_thread_supported() ) g_thread_init(NULL);
   g_assert( g_thread_supported() );
#endif

  /*
   * Process ctags output.
   */
   if ( ctagsStdin )
      ProcessCtagsFromStdin();
   else
      ProcessCtagsFromFile( pTagsFilename );

  /*
   * Create the widgets.
   */
   CreateWidgets();

   exit( 0 );
}

/******************************************************************************
 * Initialize globals (call before command line args are parsed).
 ******************************************************************************/
void InitGlobals( void )
{
  /*
   * Create the hash table for loaded modules.
   */
   pHashModules = g_hash_table_new( HashModuleKey,
                                    HashModuleCmp );
   g_assert( pHashModules );

  /*
   * Misc initializations (impractical to do a compile-time).
   */
   moduleLoadedIntoTextWidget[0] = '\0';

   widgetTagsClistWidthCol[0]	= WIDGET_TAGS_CLIST_WIDTH_COL0;
   widgetTagsClistWidthCol[1]	= WIDGET_TAGS_CLIST_WIDTH_COL1;
   widgetTagsClistWidthCol[2]	= WIDGET_TAGS_CLIST_WIDTH_COL2;
   widgetTagsClistWidthCol[3]	= WIDGET_TAGS_CLIST_WIDTH_COL3;

   widgetModulesClistWidthCol[0] = WIDGET_MODULES_CLIST_WIDTH_COL0;
   widgetModulesClistWidthCol[1] = WIDGET_MODULES_CLIST_WIDTH_COL1;

   widgetFoundClistWidthCol[0]	= WIDGET_FOUND_CLIST_WIDTH_COL0;
   widgetFoundClistWidthCol[1]	= WIDGET_FOUND_CLIST_WIDTH_COL1;
   widgetFoundClistWidthCol[2]	= WIDGET_FOUND_CLIST_WIDTH_COL2;

   findGeom[0] = 0;
   findGeom[1] = 0;
   findGeom[2] = 0;
   findGeom[3] = 0;

   mainWindowTitle[0] = '\0';

   pGStringStatusbarMsg = g_string_new( "" );

  notebookPageWithSelectedRow = NotebookPageTags();

  /*
   * If a font name of a widget is empty then its font won't be changed
   * (ie let GTK choose its font).
   */
#if 0
   strcpy( fontNameText, "" );
   strcpy( fontNameList, "" );
#else
#if 0
   /* ### THIS FONT WON'T BE DISPLAYED PROPERLY ON XFREE86 4.0.1 ### */
   strcpy( fontNameText, "-misc-fixed-medium-r-*-*-*-120-*-*-*-*-*-*" );
#else
   strcpy( fontNameText, "7x13" );
#endif
   strcpy( fontNameList, "" );
#endif

  /*
   * "-1" (invalid X) indicates to not change the geometry of the main window.
   */
   widgetMainGeom[0]		= -1;
}

/******************************************************************************
 * Parse the command line.
 *
 * GNU users have a habit of using double-dashes so the common options
 * have a GNUish variant.
 ******************************************************************************/
void ParseCommandLine( int argc, char* argv[] )
{
   int	argIdx;

   for ( argIdx = 1; argIdx < argc; ++argIdx )
   {
     /*
      * "-help"
      */
      if (    strcmp( argv[argIdx], "-help"  ) == 0
           || strcmp( argv[argIdx], "-h"     ) == 0
           || strcmp( argv[argIdx], "--help" ) == 0 )
      {
         g_print( "\"hypersrc\" by Jim Brooks jim@jimbrooks.org \n" );
         g_print( " \n" );
         g_print( "Synopsis: \n" );
         g_print( "   GTK-based program for browsing source code. \n" );
         g_print( " \n" );
         g_print( "Notice: \n" );
         g_print( "   Requires 'Exuberants Ctags' version 1.6 or greater by Darren Hiebert \n" );
         g_print( "   http://fly.hiwaay.net/~darren/ctags \n" );
         g_print( " \n" );
         g_print( "   For convenience, use the wrapper Perl script 'Hypersrc.pl' (which usually \n" );
         g_print( "   does not need arguments) from a directory containing source code. \n" );
         g_print( " \n" );
         g_print( "Options (mandatory): \n" );
         g_print( "   -ctags-x <filename> tags in Exuberant ctags -x format     \n" );
         g_print( "   OR                                                        \n" );
         g_print( "   -stdin              to pipe ctags -x output into hypersrc \n" );
         g_print( " \n" );
         g_print( "Options (important): \n" );
         g_print( "   -max-tags <num>     maximum amount of tags that can be processed \n" );
         g_print( "   -version            print version number and exit\n" );
         g_print( "   -help [or -h]       print help and exit\n" );
         g_print( " \n" );
         g_print( "Options (misc): \n" );
         g_print( "   -tab-width          tab width \n" );
         g_print( "   -geom <x y w h>     geometry of main window \n" );
         g_print( "   -column-widths <n1 n2 n3 n4> column widths \n" );
         g_print( "   -no-line-column     hide the column for which line a tag resides \n" );
         g_print( "   -font-list          font name of the list widgets \n" );
         g_print( "   -font-text          font name of the text widget \n" );
         g_print( "   -text-left          put text widget at left side \n" );
         g_print( "   -title              window title \n" );
         g_print( "   -TagModuleTypeLine  tag column layout \n" );
         g_print( "   -TagTypeModuleLine  tag column layout \n" );
         g_print( "   -ModuleTagTypeLine  tag column layout \n" );
         g_print( "   -ModuleTypeTagLine  tag column layout \n" );

         exit( 0 );
      }

     /*
      * "-ctags-x <fname>"
      *
      * This argument specifies which file holds output from ctags -x.
      */
      if ( strcmp( argv[argIdx], "-ctags-x" ) == 0 )
      {
        /*
         * Ensure another argument follows (assumed to be the filename).
         */
         if ( argIdx + 1 >= argc ) g_error( "missing argument" );

        /*
         * Set the name of the file holds output from ctags -x.
         */
         pTagsFilename = CreateString( strlen( argv[argIdx + 1] ) );
         CopyString( pTagsFilename,
                     argv[argIdx + 1],
                     strlen( argv[argIdx + 1] ) );

        /*
         * Skip the filename argument and continue.
         */
         ++argIdx;
         continue;
      }

     /*
      * "-max-tags <num>"
      *
      * This argument specifies the maximum amount of tags that can be processed.
      * An absurd maximum will be ignored.
      */
      if ( strcmp( argv[argIdx], "-max-tags" ) == 0 )
      {
        /*
         * Ensure another argument follows.
         */
         if ( argIdx + 1 >= argc ) g_error( "missing argument" );

        /*
         * Convert following arg into a binary number.
         */
#if 1
         maxTags = strtol( argv[argIdx + 1], NULL, 10 );
#else
         maxTags = g_strtod( argv[argIdx + 1], NULL );
#endif
         if ( maxTags < 1  ||  maxTags > MAX_TAGS_ABSOLUTE )
         {
            maxTags = MAX_TAGS_DEFAULT;
            SetStatusbarInitialMsg( "Specified tags maximum is unrealistic, using " MAX_TAGS_ABSOLUTE_STRING " instead." );
         }

        /*
         * Skip the following argument and continue.
         */
         ++argIdx;
         continue;
      }

     /*
      * "-tab-width <num>"
      *
      * This argument specifies the tab width.
      */
      if (    strcmp( argv[argIdx], "-tab-width" ) == 0
           || strcmp( argv[argIdx], "-tab" )       == 0 )
      {
        /*
         * Ensure another argument follows.
         */
         if ( argIdx + 1 >= argc ) g_error( "missing argument" );

        /*
         * Convert following arg into a binary number.
         */
         cmdlineTabWidth = strtol( argv[argIdx + 1], NULL, 10 );

        /*
         * If the specified tab width is unreasonable,
         * then assign zero which means to use the GTK default tab width.
         */
         if ( cmdlineTabWidth < 1  ||  cmdlineTabWidth > 100 )
            cmdlineTabWidth = 0;

        /*
         * Skip the following argument and continue.
         */
         ++argIdx;
         continue;
      }

     /*
      * "-geom <x y w h>"
      *
      * This argument specifies the geometry of the main window.
      */
      if ( strcmp( argv[argIdx], "-geom" ) == 0 )
      {
         int i;

        /*
         * Ensure at least 4 arguments follow.
         */
         if ( argIdx + 4 >= argc ) g_error( "missing argument" );

        /*
         * Parse the geom.
         */
         for ( i = 0; i < 4; ++i )
         {
           /*
            * Try to convert this arg into a binary number.
            */
            widgetMainGeom[i] = strtol( argv[argIdx + 1], NULL, 10 );

           /*
            * Ensure this geom component is reasonable.
            */
            if (     i <= 1  /* or 0 */
                 &&  (widgetMainGeom[i] < 0  ||  widgetMainGeom[i] > MAX_GEOM_X_Y) )
            {
               g_error( "-geom arg is unrealistic" );
            }
            if (     i >= 2  /* or 3 */
                 &&  (widgetMainGeom[i] < 0  ||  widgetMainGeom[i] > MAX_GEOM_W_H) )
            {
               g_error( "-geom arg is unrealistic" );
            }

           /*
            * To the next arg.
            */
            ++argIdx;
         }

         continue;
      }

     /*
      * "-column-widths <n1 n2 n3 n4>"
      *
      * This argument specifies the width of the tag columns.
      */
      if ( strcmp( argv[argIdx], "-column-widths" ) == 0 )
      {
         int i;

        /*
         * Ensure at least 4 arguments follow.
         */
         if ( argIdx + 4 >= argc ) g_error( "missing argument" );

        /*
         * Parse the column widths.
         */
         for ( i = 0; i < WIDGET_TAGS_CLIST_COLUMNS; ++i )
         {
           /*
            * Try to convert this arg into a binary number.
            */
            widgetTagsClistWidthCol[i] = strtol( argv[argIdx + 1], NULL, 10 );

           /*
            * Ensure the width is reasonable.
            */
            if ( widgetTagsClistWidthCol[i] <= MIN_WIDGET_TAGS_CLIST_WIDTH_COL )
            {
               widgetTagsClistWidthCol[i] = MIN_WIDGET_TAGS_CLIST_WIDTH_COL;
            }

           /*
            * To the next arg.
            */
            ++argIdx;
         }

         continue;
      }

     /*
      * "-no-line-column"
      */
      if ( strcmp( argv[argIdx], "-no-line-column" ) == 0 )
      {
         hideTagLineColumn = TRUE;
         continue;
      }

     /*
      * "-TagModuleTypeLine"
      */
      if ( strcmp( argv[argIdx], "-TagModuleTypeLine" ) == 0 )
      {
         tagsColumnsLayout = TAG_COL_LAYOUT_TAG_MODULE_TYPE_LINE;
         continue;
      }

     /*
      * "-TagTypeModuleLine"
      */
      if ( strcmp( argv[argIdx], "-TagTypeModuleLine" ) == 0 )
      {
         tagsColumnsLayout = TAG_COL_LAYOUT_TAG_TYPE_MODULE_LINE;
         continue;
      }

     /*
      * "-ModuleTagTypeLine"
      */
      if ( strcmp( argv[argIdx], "-ModuleTagTypeLine" ) == 0 )
      {
         tagsColumnsLayout = TAG_COL_LAYOUT_MODULE_TAG_TYPE_LINE;
         continue;
      }

     /*
      * "-ModuleTypeTagLine"
      */
      if ( strcmp( argv[argIdx], "-ModuleTypeTagLine" ) == 0 )
      {
         tagsColumnsLayout = TAG_COL_LAYOUT_MODULE_TYPE_TAG_LINE;
         continue;
      }

     /*
      * "-font-list <font>"
      * "-font-text <font>"
      *
      * These arguments specify the fonts used for clist and text widgets.
      */
      if (    strcmp( argv[argIdx], "-font-list" ) == 0
           || strcmp( argv[argIdx], "-font-text" ) == 0 )
      {
        /*
         * Ensure another argument follows.
         */
         if ( argIdx + 1 >= argc ) g_error( "missing argument" );

        /*
         * Record the font name in the appropriate var.
         */
         if ( strcmp( argv[argIdx], "-font-list"  ) == 0 )
            strcpy( fontNameList, argv[argIdx + 1] );
         else
            strcpy( fontNameText, argv[argIdx + 1] );

        /*
         * Skip the following argument and continue.
         */
         ++argIdx;
         continue;
      }

     /*
      * "-color-x-fg <r g b>"
      * "-color-x-bg <r g b>"
      *
      * These arguments specify the colors used for clist and text widgets.
      */
      if (    strcmp( argv[argIdx], "-color-fg-list" ) == 0
           || strcmp( argv[argIdx], "-color-bg-list" ) == 0
           || strcmp( argv[argIdx], "-color-fg-text" ) == 0
           || strcmp( argv[argIdx], "-color-bg-text" ) == 0 )
      {
         int	  i;
         long     colorPart;
         GdkColor colorRgb;

        /*
         * Ensure at least 3 arguments follow.
         */
         if ( argIdx + 3 >= argc ) g_error( "missing argument" );

        /*
         * Parse the colors.
         */
         for ( i = 0; i < 3; ++i )
         {
           /*
            * Try to convert this arg into a binary number.
            */
            colorPart = strtol( argv[argIdx + 1], NULL, 16 );

           /*
            * Ensure the color is reasonable.
            * TODO
            */

           /*
            * Store the parsed color in the appropriate RGB elem.
            */
            switch ( i )
            {
               case 0:
                  colorRgb.red   = colorPart; break;
               case 1:
                  colorRgb.green = colorPart; break;
               case 2:
                  colorRgb.blue  = colorPart; break;
            }

           /*
            * To the next arg.
            */
            ++argIdx;

           /*
            * When all three color parts have been parsed,
            * assign the GdkColor array of the specified widget.
            */
            if ( i == 2 )
            {
               if ( strcmp( argv[argIdx - 3], "-color-fg-list" ) == 0 )
                  colorClistWidgetTextFg = colorRgb;
               else
               if ( strcmp( argv[argIdx - 3], "-color-bg-list" ) == 0 )
                  colorClistWidgetTextBg = colorRgb;
               else
               if ( strcmp( argv[argIdx - 3], "-color-fg-text" ) == 0 )
                  colorTextWidgetTextFg = colorRgb;
               else
               if ( strcmp( argv[argIdx - 3], "-color-bg-text" ) == 0 )
                  colorTextWidgetTextBg = colorRgb;
            }
         }
         continue;
      }

     /*
      * "-text-left"
      *
      * If present then put the text widget at the left side.
      */
      if ( strcmp( argv[argIdx], "-text-left" ) == 0 )
      {
         textLeft = TRUE;
         continue;
      }

     /*
      * "-title"
      */
      if ( strcmp( argv[argIdx], "-title" ) == 0 )
      {
        /*
         * Ensure another argument follows.
         */
         if ( argIdx + 1 >= argc ) g_error( "missing argument" );

        /*
         * Record the passed window title.
         */
         strcpy( mainWindowTitle, argv[argIdx + 1] );

        /*
         * Skip the following argument and continue.
         */
         ++argIdx;
         continue;
      }

     /*
      * "-version"
      */
      if (    strcmp( argv[argIdx], "-version"  ) == 0
           || strcmp( argv[argIdx], "--version" ) == 0 )
      {
         g_print( "%s\n", HYPERSRC_VERSION_STRING );
         exit( 0 );
      }

     /*
      * "-stdin"
      */
      if ( strcmp( argv[argIdx], "-stdin"  ) == 0 )
      {
         ctagsStdin = TRUE;
         continue;
      }

     /*
      * Oops, arg was not recognized.
      */
      g_error( "Unrecognized command line argument (use -help)." );
   }

   return;
}
